@@ -2,10 +2,8 @@
 # this repository contains the full copyright notices and license terms.
 
 from trytond.pool import Pool
-from . import ir
-from . import work
-from . import timesheet
-from . import party
+
+from . import ir, party, timesheet, work
 
 
 def register():
