@@ -2,17 +2,16 @@
 # this repository contains the full copyright notices and license terms.
 
 import datetime
-
-from sql import Null
+from collections import defaultdict
 
 from trytond.cache import Cache
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelView, ModelSQL, fields, sequence_ordered, tree, DeactivableMixin)
-from trytond.transaction import Transaction
+    DeactivableMixin, Index, ModelSQL, ModelView, fields, sequence_ordered,
+    sum_tree, tree)
 from trytond.pool import Pool
-from trytond.pyson import Eval, If, Bool, PYSONEncoder
-from trytond.tools import reduce_ids, grouped_slice
+from trytond.pyson import Bool, Eval, If, PYSONEncoder, TimeDelta
+from trytond.transaction import Transaction, without_check_access
 
 from .exceptions import WorkProgressValidationError
 
@@ -121,14 +120,13 @@ class WorkStatus(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
 class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
     'Work Effort'
     __name__ = 'project.work'
-    name = fields.Char('Name', required=True, select=True)
+    name = fields.Char("Name", required=True)
     type = fields.Selection([
             ('project', 'Project'),
             ('task', 'Task')
             ],
-        'Type', required=True, select=True)
-    company = fields.Many2One('company.company', 'Company', required=True,
-        select=True)
+        "Type", required=True)
+    company = fields.Many2One('company.company', "Company", required=True)
     party = fields.Many2One('party.party', 'Party',
         states={
             'invisible': Eval('type') != 'project',
@@ -136,34 +134,38 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         context={
             'company': Eval('company', -1),
             },
-        depends=['type', 'company'])
+        depends={'company'})
     party_address = fields.Many2One('party.address', 'Contact Address',
         domain=[('party', '=', Eval('party'))],
         states={
             'invisible': Eval('type') != 'project',
-            }, depends=['party', 'type'])
+            })
     timesheet_works = fields.One2Many(
         'timesheet.work', 'origin', 'Timesheet Works', readonly=True, size=1)
-    timesheet_available = fields.Function(
-        fields.Boolean('Available on timesheets'),
+    timesheet_available = fields.Function(fields.Boolean(
+            "Available on timesheets",
+            help="Check to record time spent."),
         'get_timesheet_available', setter='set_timesheet_available')
     timesheet_start_date = fields.Function(fields.Date('Timesheet Start',
             states={
                 'invisible': ~Eval('timesheet_available'),
-                },
-            depends=['timesheet_available']),
+                }),
         'get_timesheet_date', setter='set_timesheet_date')
     timesheet_end_date = fields.Function(fields.Date('Timesheet End',
             states={
                 'invisible': ~Eval('timesheet_available'),
-                },
-            depends=['timesheet_available']),
+                }),
         'get_timesheet_date', setter='set_timesheet_date')
     timesheet_duration = fields.Function(fields.TimeDelta('Duration',
             'company_work_time',
             help="Total time spent on this work and the sub-works."),
         'get_total')
-    effort_duration = fields.TimeDelta('Effort', 'company_work_time',
+    effort_duration = fields.TimeDelta(
+        "Effort", 'company_work_time',
+        domain=['OR',
+            ('effort_duration', '=', None),
+            ('effort_duration', '>=', TimeDelta()),
+            ],
         help="Estimated Effort for this work.")
     total_effort = fields.Function(fields.TimeDelta('Total Effort',
             'company_work_time',
@@ -187,23 +189,28 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
                 }),
         'get_total')
     comment = fields.Text('Comment')
-    parent = fields.Many2One('project.work', 'Parent',
-        left='left', right='right', ondelete='RESTRICT',
+    parent = fields.Many2One(
+        'project.work', 'Parent', path='path', ondelete='RESTRICT',
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
-    left = fields.Integer('Left', required=True, select=True)
-    right = fields.Integer('Right', required=True, select=True)
+            ])
+    path = fields.Char("Path")
     children = fields.One2Many('project.work', 'parent', 'Children',
         domain=[
             ('company', '=', Eval('company', -1)),
-            ],
-        depends=['company'])
+            ])
     status = fields.Many2One(
-        'project.work.status', "Status", required=True, select=True,
-        domain=[If(Bool(Eval('type')), ('types', 'in', Eval('type')), ())],
-        depends=['type'])
+        'project.work.status', "Status", required=True,
+        domain=[If(Bool(Eval('type')), ('types', 'in', Eval('type')), ())])
+
+    @classmethod
+    def __setup__(cls):
+        cls.path.search_unaccented = False
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.path, Index.Similarity(begin=True))),
+                })
 
     @staticmethod
     def default_type():
@@ -219,97 +226,19 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         WorkStatus = pool.get('project.work.status')
         return WorkStatus.get_default_status(cls.default_type())
 
-    @classmethod
-    def default_left(cls):
-        return 0
-
-    @classmethod
-    def default_right(cls):
-        return 0
-
     @classmethod
     def __register__(cls, module_name):
-        TimesheetWork = Pool().get('timesheet.work')
-        transaction = Transaction()
-        cursor = transaction.connection.cursor()
-        update = transaction.connection.cursor()
         table_project_work = cls.__table_handler__(module_name)
-        project = cls.__table__()
-        timesheet = TimesheetWork.__table__()
-
-        work_exist = table_project_work.column_exist('work')
-        add_parent = (not table_project_work.column_exist('parent')
-            and work_exist)
-        add_company = (not table_project_work.column_exist('company')
-            and work_exist)
-        add_name = (not table_project_work.column_exist('name')
-            and work_exist)
 
         super(Work, cls).__register__(module_name)
 
-        # Migration from 3.4: change effort into timedelta effort_duration
-        if table_project_work.column_exist('effort'):
-            cursor.execute(*project.select(project.id, project.effort,
-                    where=project.effort != Null))
-            for id_, effort in cursor:
-                duration = datetime.timedelta(hours=effort)
-                update.execute(*project.update(
-                        [project.effort_duration],
-                        [duration],
-                        where=project.id == id_))
-            table_project_work.drop_column('effort')
-
-        # Migration from 3.6: add parent, company, drop required on work,
-        # fill name
-        if add_parent:
-            second_project = cls.__table__()
-            query = project.join(timesheet,
-                condition=project.work == timesheet.id
-                ).join(second_project,
-                    condition=timesheet.parent == second_project.work
-                    ).select(project.id, second_project.id)
-            cursor.execute(*query)
-            for id_, parent in cursor:
-                update.execute(*project.update(
-                        [project.parent],
-                        [parent],
-                        where=project.id == id_))
-            cls._rebuild_tree('parent', None, 0)
-        if add_company:
-            cursor.execute(*project.join(timesheet,
-                    condition=project.work == timesheet.id
-                    ).select(project.id, timesheet.company))
-            for id_, company in cursor:
-                update.execute(*project.update(
-                        [project.company],
-                        [company],
-                        where=project.id == id_))
-        table_project_work.not_null_action('work', action='remove')
-        if add_name:
-            cursor.execute(*project.join(timesheet,
-                    condition=project.work == timesheet.id
-                    ).select(project.id, timesheet.name))
-            for id_, name in cursor:
-                update.execute(*project.update(
-                        [project.name],
-                        [name],
-                        where=project.id == id_))
-
-        # Migration from 4.0: remove work
-        if work_exist:
-            table_project_work.drop_constraint('work_uniq')
-            cursor.execute(*project.select(project.id, project.work,
-                    where=project.work != Null))
-            for project_id, work_id in cursor:
-                update.execute(*timesheet.update(
-                        [timesheet.origin, timesheet.name],
-                        ['%s,%s' % (cls.__name__, project_id), Null],
-                        where=timesheet.id == work_id))
-            table_project_work.drop_column('work')
-
         # Migration from 5.4: replace state by status
         table_project_work.not_null_action('state', action='remove')
 
+        # Migration from 6.0: remove left and right
+        table_project_work.drop_column('left')
+        table_project_work.drop_column('right')
+
     @fields.depends('type', 'status')
     def on_change_type(self):
         pool = Pool()
@@ -429,42 +358,11 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
                     name: value,
                     })
 
-    @classmethod
-    def sum_tree(cls, works, values, parents):
-        result = values.copy()
-        works = set((w.id for w in works))
-        leafs = works - set(parents.values())
-        while leafs:
-            for work in leafs:
-                works.remove(work)
-                parent = parents.get(work)
-                if parent in result:
-                    result[parent] += result[work]
-            next_leafs = set(works)
-            for work in works:
-                parent = parents.get(work)
-                if not parent:
-                    continue
-                if parent in next_leafs and parent in works:
-                    next_leafs.remove(parent)
-            leafs = next_leafs
-        return result
-
     @classmethod
     def get_total(cls, works, names):
-        cursor = Transaction().connection.cursor()
-        table = cls.__table__()
-
         works = cls.search([
                 ('parent', 'child_of', [w.id for w in works]),
                 ])
-        work_ids = [w.id for w in works]
-        parents = {}
-        for sub_ids in grouped_slice(work_ids):
-            where = reduce_ids(table.id, sub_ids)
-            cursor.execute(*table.select(table.id, table.parent,
-                    where=where))
-            parents.update(cursor)
 
         if 'total_progress' in names and 'total_effort' not in names:
             names = list(names)
@@ -473,7 +371,7 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         result = {}
         for name in names:
             values = getattr(cls, '_get_%s' % name)(works)
-            result[name] = cls.sum_tree(works, values, parents)
+            result[name] = sum_tree(works, values)
 
         if 'total_progress' in names:
             digits = cls.total_progress.digits[1]
@@ -490,11 +388,13 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
 
     @classmethod
     def _get_total_effort(cls, works):
-        return {w.id: w.effort_duration or datetime.timedelta() for w in works}
+        return defaultdict(
+            datetime.timedelta,
+            {w.id: w.effort_duration or datetime.timedelta() for w in works})
 
     @classmethod
     def _get_timesheet_duration(cls, works):
-        durations = {}
+        durations = defaultdict(datetime.timedelta)
         for work in works:
             value = datetime.timedelta()
             for timesheet_work in work.timesheet_works:
@@ -505,7 +405,9 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
 
     @classmethod
     def _get_total_progress(cls, works):
-        return {w.id: w.effort_hours * (w.progress or 0) for w in works}
+        return defaultdict(
+            int,
+            {w.id: w.effort_hours * (w.progress or 0) for w in works})
 
     @classmethod
     def copy(cls, project_works, default=None):
@@ -515,11 +417,20 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
             default = {}
         else:
             default = default.copy()
-        default.setdefault('children', None)
         default.setdefault('progress', None)
         default.setdefault(
             'status', lambda data: WorkStatus.get_default_status(data['type']))
-        return super().copy(project_works, default=default)
+        new_works = super().copy(project_works, default=default)
+        to_save = []
+        for work, new_work in zip(project_works, new_works):
+            if work.timesheet_available:
+                new_work.timesheet_available = work.timesheet_available
+                new_work.timesheet_start_date = work.timesheet_start_date
+                new_work.timesheet_end_date = work.timesheet_end_date
+                to_save.append(new_work)
+        if to_save:
+            cls.save(to_save)
+        return new_works
 
     @classmethod
     def delete(cls, project_works):
@@ -532,7 +443,7 @@ class Work(sequence_ordered(), tree(separator='\\'), ModelSQL, ModelView):
         super(Work, cls).delete(project_works)
 
         if timesheet_works:
-            with Transaction().set_context(_check_access=False):
+            with without_check_access():
                 TimesheetWork.delete(timesheet_works)
 
     @classmethod
